@@ -536,7 +536,7 @@ static gboolean get_xfer_progress(const void *buf, gsize len,
 	struct obc_transfer *transfer = user_data;
 
 	if (transfer->fd > 0) {
-		gint w;
+		int w;
 
 		w = write(transfer->fd, buf, len);
 		if (w < 0)
@@ -596,8 +596,8 @@ static void get_xfer_progress_first(GObex *obex, GError *err, GObexPacket *rsp,
 
 	rspcode = g_obex_packet_get_operation(rsp, &final);
 	if (rspcode != G_OBEX_RSP_SUCCESS && rspcode != G_OBEX_RSP_CONTINUE) {
-		err = g_error_new(OBC_TRANSFER_ERROR, rspcode,
-					"Transfer failed (0x%02x)", rspcode);
+		err = g_error_new(OBC_TRANSFER_ERROR, rspcode, "%s",
+						g_obex_strerror(rspcode));
 		xfer_complete(obex, err, transfer);
 		g_error_free(err);
 		return;
